package edu.qrg.vmodel;

import edu.qrg.utils.*;
import edu.qrg.utils.CMLio.CMLable;
import edu.qrg.utils.PList.PLinterface;
import edu.qrg.utils.XMLio.XMLable;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.ClassNotFoundException;
import java.lang.Cloneable;
import java.lang.Math;
import java.lang.Object;
import java.lang.String;
import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;
import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

/** <PRE> 
* Sun Java 1.3.1 -Sourcepath QRG/VModel/development/code/library/ontologies
* ---------------------------------------------------------------------------
*  2003 Northwestern University Qualitative Reasoning Group Kenneth D. Forbus
* ---------------------------------------------------------------------------
* File name: Ontology.java
*    System: VModel
*     Since: v0.6
*    Author: Leo C. Ureel II
* ---------------------------------------------------------------------------
*   Class: Ontology
* Purpose: Defines an ontology of elements. In VModel there is a single
*          ontology object that is used to maintain canonical ontological
*          elements.  All other ontological elements are contained in
*          libraries and are based on those in the ontology.
* ---------------------------------------------------------------------------
* Notes:
* * Should the file IO routines be included in this class?
* ---------------------------------------------------------------------------
* </PRE> */
public class Ontology extends PList
		implements Cloneable, Serializable, Ontology.OntInterface {
	// Constructors
	// -----------------------------------------------------------------------
	/** General initialization. */
	public Ontology () {
		id(composeUniqueIdentifier());
		allOntologies().addElement(this);
    } // default constructor
		
	/** Initializes most frequently used slots. */
    public Ontology(String id, String name, String documentation, String helpURL) {
    	this();
    	id(id);
    	name(name);
    	documentation(documentation);
    	helpURL(helpURL);
    } // constructor

	// Interfaces
	// -----------------------------------------------------------------------
	/** <PRE>
	* ------------------------------------------------------------------------
	* Interface: Ontology.OntInterface
	* Purpose: In general, anything implementing this interface should be
	*          acceptable as a Ontology.
	* ------------------------------------------------------------------------
	* </PRE> 
	*/
	public interface OntInterface extends Cloneable, Serializable, VModelConstants, 
			PList.PLinterface, CMLio.CMLable, XMLio.XMLable, Ontology.Element, Ontology.Category {
		// Properties & Accessors
		// --------------------------------------------------------------------
		/** List of ontological elements contained in the Ontology.
		    Note: Ontological elements are assumed type Ontology.Element. */
		public Vector elements ();

		/** Return the unique string identifier for the object. */
		public String id ();
		/** Set & return the unique string identifier for the object. */
		public String id (String s);

		/** Keeps track of the selected element in the tree model.
		* * Is this still being used?
		*/
		public String selectedElement ();
		/** @see #selectedElement() */
		public String selectedElement (String definitionID);
	    public void reload(TreeNode node);

		// Methods
		// --------------------------------------------------------------------
		/** Add an ontological element to the ontology. */
		public void add (Ontology.Element a);
		/** Remove an ontological element from the ontology. */
		public void remove (Ontology.Element a);

		/** Find an ontological element by id.
			 Searches categories. 
		    Returns null if not found. */
		public Ontology.Element findElement (String id);

		/** True if ontology contains an ontological element having id.
			 Searches categories. 
		    Returns false if not found. */
		public boolean contains (String id);

		/** Saves the ontology to filename. */
		public void save (String filename);
		
		/* Load should be in here, but java doesn't allow static members
		   to be placed in interfaces.  Grrrr. */
		/* Recursively list all children */
		public Vector flatten ();
	} // interface
	
	/** <PRE>
	* ------------------------------------------------------------------------
	* Interface: Ontology.Element
	* Purpose: In general, anything implementing this interface may be
	*          add to an Ontology as an ontological element.
	* ------------------------------------------------------------------------
	* </PRE> 
	*/
	public interface Element 
			extends Cloneable, Serializable, VModelConstants, PList.PLinterface, TreeNode {
		// Properties & Accessors
		// --------------------------------------------------------------------
		/** Return the unique string identifier for the object. */
		public String id ();
		/** Set & return the unique string identifier for the object. */
		public String id (String s);

		/** Return a string name for the object. */
		public String name ();
		/** Set & return a string name for the object. */
		public String name (String s);

		/** Return a documentation string for the object. */
		public String documentation ();
		/** Sets & return a documentation string for the object. */
		public String documentation (String s);
		
		/** Return a string filename pointing to a web-based
		    documentation file for the object. */
		public String helpURL ();
		/** Sets & returns a string filename pointing to a web-based
		    documentation file for the object. */
		public String helpURL (String s);
		
		/** Returns a string filename pointing to a gif or jpg file.
		    Filename does not have to include file-extension. */
		public String icon ();
		/** Sets * returns a string filename pointing to a gif or jpg file.
		    Filename does not have to include file-extension. */
		public String icon (String name);
		
		/** Id of category to which the ontological element belongs. */
		public String category ();
		/** @see #category() */
		public String category (String s);

		/** Type of ontological element. */
		public String type ();
		/** @see #type() */
		public String type (String s);
	
		/** Type of ontological element. */
		public String typeName ();
		/** @see #type() */
		public String typeName (String s);
	
		public boolean typep (String typeID);

		/** Represent object as string. */
		public String stringify ();
		/** Represent object as string with indentation. */
		public String stringify (int indent);
	} // interface
	
	/** <PRE>
	* ------------------------------------------------------------------------
	* Interface: Ontology.Category
	* Purpose: In general, anything implementing this interface may be
	*          add to an Ontology as an ontological element.
	* ------------------------------------------------------------------------
	* </PRE> 
	*/
	public interface Category 
			extends Cloneable, Serializable, VModelConstants, PList.PLinterface,
				Ontology.Element, TreeModel {
		/** Returns a string filename pointing to a gif or jpg file.
			 Filename is the same as returned by icon() but
			 "Open" is appended if expanded and "Closed" otherwise.
		    Filename does not include file-extension. */
		public String icon (boolean expanded);
						
	} // interface
	
	// Properties & Accessors
	// -----------------------------------------------------------------------
	/** @serial */
	protected String selectedElement = null;
	/** Stashes the id of the ontological element currently
	*   selected in the tree model.
	*/
	public String selectedElement () { return(selectedElement); }
	/** @see #selectedElement() */
	public String selectedElement (String definitionID) { 
		return(selectedElement = definitionID); 
	} // writer
	
	/** Purpose: Returns the string id of the category to which this
	             category belongs. */
	public String category () { return(getString(CATEGORY)); }
	/** @see #category() */
	public String category (String s) { return((String) set(CATEGORY, s)); }
	
	/** Purpose: Returns the string filename of an image representing
	             the library. */
	public String icon () { return(getString(ICON)); }
	/** @see #icon() */
	public String icon (String s) { return((String) set(ICON, s)); }
	/** Purpose: Returns the string filename of an image representing
	             the library tree node in an Open or Close state. */
	public String icon (boolean expanded) { 
		String result = icon();
		if (expanded) result += "Open";
		else result += "Closed";
		return(result); 
	} // method
	
	/** Type of ontological element. */
	public String type () { return(LIBRARY); }
	/** @see #type() */
	public String type (String s) { return(LIBRARY); }

	/** Type of ontological element. */
	public String typeName () { return(LIBRARY); }
	/** @see #type() */
	public String typeName (String s) { return(LIBRARY); }

	/** Type check */
	public boolean typep (String typeID) {
		boolean result = false;
		Definition mumble = null;
		result = id().equalsIgnoreCase(typeID) ||
			name().equalsIgnoreCase(typeID) ||
			category().equalsIgnoreCase(typeID) ||
			type().equalsIgnoreCase(typeID);
		return(result);
	} // method
	
	// TreeModel Methods
	// -----------------------------------------------------------------------
	/** Use accessors. */
	protected transient Vector treeModelListeners = null;
	/** Stash for listeners. Default is new Vector(). */
	public Vector treeModelListeners () {
		if (treeModelListeners == null) treeModelListeners = new Vector(); 
		return(treeModelListeners); 
	} // reader
	
	/** Add a tree model listener.  Wow! These are exciting comments! */
	public void addTreeModelListener(TreeModelListener l) {
		treeModelListeners().addElement(l);
	} // method
	
	/** Remove tree model listener. */ 
	public void removeTreeModelListener(TreeModelListener l) {
		treeModelListeners().removeElement(l);
	} // method
	
	/** Returns child tree node. */
	public Object getChild(Object category, int index) {
		return(((TreeNode) category).getChildAt(index));
	} // method
	
	/** Returns number of child nodes. */
	public int getChildCount(Object category) {
		return(((TreeNode) category).getChildCount());
	} // method
	
	/** Returns location of child node in tree. */
	public int getIndexOfChild(Object category, Object child) {
		return(((TreeNode) category).getIndex((TreeNode) child));
	} // method
	
	/** Return root node of tree. */
	public Object getRoot() { return(this); } // method
	
	/** Return true if node has no children. */
	public boolean isLeaf(Object node) {
		return(((TreeNode) node).isLeaf());
	} // method
	
	/** Does this ever get called?  How can we use it to allow renaming of
	    ontological elements in the tree display? */
	public void valueForPathChanged(TreePath path, Object newValue) {
		System.out.println("Ontology.valueForPathChanged path=" + path + " newValue=" + newValue);
	} // method

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance 
     * is lazily created using the parameters passed into 
     * the fire method.
     * @see EventListenerList
     */
    protected void fireTreeNodesChanged(Object source, Object[] path, 
                                        int[] childIndices, 
                                        Object[] children) {
       	TreeModelEvent e = null;
       	TreeModelListener l = null;
		for (Enumeration enum = treeModelListeners().elements(); enum.hasMoreElements();) {
			l = (TreeModelListener) enum.nextElement();
                if (e == null)
                    e = new TreeModelEvent(source, path, 
                                           childIndices, children);
                l.treeNodesChanged(e);
        }
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance 
     * is lazily created using the parameters passed into 
     * the fire method.
     * @see EventListenerList
     */
    protected void fireTreeNodesInserted(Object source, Object[] path, 
                                        int[] childIndices, 
                                        Object[] children) {
       	TreeModelEvent e = null;
       	TreeModelListener l = null;
		for (Enumeration enum = treeModelListeners().elements(); enum.hasMoreElements();) {
			l = (TreeModelListener) enum.nextElement();
                // Lazily create the event:
                if (e == null)
                    e = new TreeModelEvent(source, path, 
                                           childIndices, children);
                l.treeNodesInserted(e);
        }
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance 
     * is lazily created using the parameters passed into 
     * the fire method.
     * @see EventListenerList
     */
    protected void fireTreeNodesRemoved(Object source, Object[] path, 
                                        int[] childIndices, 
                                        Object[] children) {
       	TreeModelEvent e = null;
       	TreeModelListener l = null;
		for (Enumeration enum = treeModelListeners().elements(); enum.hasMoreElements();) {
			l = (TreeModelListener) enum.nextElement();
                if (e == null)
                    e = new TreeModelEvent(source, path, 
                                           childIndices, children);
                l.treeNodesRemoved(e);
        }
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance 
     * is lazily created using the parameters passed into 
     * the fire method.
     * @see EventListenerList
     */
    protected void fireTreeStructureChanged(Object source, Object[] path, 
                                        int[] childIndices, 
                                        Object[] children) {
       	TreeModelEvent e = null;
       	TreeModelListener l = null;
		for (Enumeration enum = treeModelListeners().elements(); enum.hasMoreElements();) {
			l = (TreeModelListener) enum.nextElement();
                if (e == null)
                    e = new TreeModelEvent(source, path, 
                                           childIndices, children);
                l.treeStructureChanged(e);
        }
    }

    /**
      * Invoke this method after you've changed how node is to be
      * represented in the tree.
      */
 //   public void nodeChanged(TreeNode node) {
 //       if(listenerList != null && node != null) {
 //           TreeNode         parent = node.getParent();
//
//            if(parent != null) {
//                int        anIndex = parent.getIndex(node);
//                if(anIndex != -1) {
//                    int[]        cIndexs = new int[1];
//
//                    cIndexs[0] = anIndex;
//                   nodesChanged(parent, cIndexs);
//                }
//            }
//	    else if (node == getRoot()) {
//		nodesChanged(node, null);
//	    }
//        }
//    }

    /**
     * Invoke this method if you've modified the TreeNodes upon which this
     * model depends.  The model will notify all of its listeners that the
     * model has changed below the node <code>node</code> (PENDING).
     */
    public void reload(TreeNode node) {
        if(node != null) {
            fireTreeStructureChanged(this, getPathToRoot(node), null, null);
        }
    }

    /**
      * Invoke this method after you've inserted some TreeNodes into
      * node.  childIndices should be the index of the new elements and
      * must be sorted in ascending order.
      */
//    public void nodesWereInserted(TreeNode node, int[] childIndices) {
//        if(listenerList != null && node != null && childIndices != null
//           && childIndices.length > 0) {
//            int               cCount = childIndices.length;
//            Object[]          newChildren = new Object[cCount];
//
//            for(int counter = 0; counter < cCount; counter++)
//                newChildren[counter] = node.getChildAt(childIndices[counter]);
//            fireTreeNodesInserted(this, getPathToRoot(node), childIndices, 
//                                  newChildren);
//        }
//    }
    
    /**
      * Invoke this method after you've removed some TreeNodes from
      * node.  childIndices should be the index of the removed elements and
      * must be sorted in ascending order. And removedChildren should be
      * the array of the children objects that were removed.
      */
    public void nodesWereRemoved(TreeNode node, int[] childIndices,
                                 Object[] removedChildren) {
        if(node != null && childIndices != null) {
            fireTreeNodesRemoved(this, getPathToRoot(node), childIndices, 
                                 removedChildren);
        }
    }

    /**
      * Invoke this method after you've changed how the children identified by
      * childIndicies are to be represented in the tree.
      */
    public void nodesChanged(TreeNode node, int[] childIndices) {
        if(node != null) {
	    if (childIndices != null) {
		int            cCount = childIndices.length;

		if(cCount > 0) {
		    Object[]       cChildren = new Object[cCount];

		    for(int counter = 0; counter < cCount; counter++)
			cChildren[counter] = node.getChildAt
			    (childIndices[counter]);
		    fireTreeNodesChanged(this, getPathToRoot(node),
					 childIndices, cChildren);
		}
	    }
	    else if (node == getRoot()) {
		fireTreeNodesChanged(this, getPathToRoot(node), null, null);
	    }
        }
    }

    /**
      * Invoke this method if you've totally changed the children of
      * node and its childrens children...  This will post a
      * treeStructureChanged event.
      */
    public void nodeStructureChanged(TreeNode node) {
        if(node != null) {
           fireTreeStructureChanged(this, getPathToRoot(node), null, null);
        }
    }

    /**
     * Builds the parents of node up to and including the root node,
     * where the original node is the last element in the returned array.
     * The length of the returned array gives the node's depth in the
     * tree.
     * 
     * @param aNode the TreeNode to get the path for
     * @param an array of TreeNodes giving the path from the root to the
     *        specified node. 
     */
    public TreeNode[] getPathToRoot(TreeNode aNode) {
        return getPathToRoot(aNode, 0);
    }

    /**
     * Builds the parents of node up to and including the root node,
     * where the original node is the last element in the returned array.
     * The length of the returned array gives the node's depth in the
     * tree.
     * 
     * @param aNode  the TreeNode to get the path for
     * @param depth  an int giving the number of steps already taken towards
     *        the root (on recursive calls), used to size the returned array
     * @return an array of TreeNodes giving the path from the root to the
     *         specified node 
     */
    protected TreeNode[] getPathToRoot(TreeNode aNode, int depth) {
        TreeNode[]              retNodes;
	// This method recurses, traversing towards the root in order
	// size the array. On the way back, it fills in the nodes,
	// starting from the root and working back to the original node.

        /* Check for null, in case someone passed in a null node, or
           they passed in an element that isn't rooted at root. */
        if(aNode == null) {
            if(depth == 0)
                return null;
            else
                retNodes = new TreeNode[depth];
        }
        else {
            depth++;
            if(aNode == getRoot())
                retNodes = new TreeNode[depth];
            else
                retNodes = getPathToRoot(aNode.getParent(), depth);
            retNodes[retNodes.length - depth] = aNode;
        }
        return retNodes;
    }

	// TreeNode Methods
	// -----------------------------------------------------------------------
	/** Enumerate child nodes. */
	public Enumeration children() {
		return(elements().elements());
	} // method
	
	/** Yes, allow child nodes. */
	public boolean getAllowsChildren() {
		return(true);
	} // method
	
	/** Return child node at index. */
	public TreeNode getChildAt(int childIndex) {
		TreeNode result = (TreeNode) elements().elementAt(childIndex);
		return(result);
	} // method
	
	/** Return count of child nodes. */
	public int getChildCount() {
		int result = elements().size();
		return(result);
	} // method
	
	/** Return the location of the specified node. */
	public int getIndex(TreeNode node) {
		return(elements().indexOf(node));
	} // method
	
	/** Return the parent category. */
	public TreeNode getParent() {
		Object result = null;
		if (category() != null) result = find(category());
		if (! (result instanceof TreeNode)) result = null;
		return((TreeNode) result);
	} // method
	
	/** true if this node a leaf? */
	public boolean isLeaf() {
		boolean result = elements().isEmpty();
		return(result);
	} // method

	// Properties & Accessors
	// -----------------------------------------------------------------------
	/** Counter of the number of ontology objects created in session. */
	protected static long sessionCount = 0;

	/** String identifier for object. */
	public String id () { return(getString(ID)); }
	/** @see #id() */
	public String id (String s) { return((String) set(ID, s)); }
	
	/** String label for object.
	* * May contain spaces.
	* * Doesn't have to be unique.
	*/
	public String name () { return(getString(NAME)); }
	/** @see #name() */
	public String name (String s) { return((String) set(NAME, s)); }
	
	/** Documentation string for object. */
	public String documentation () { return(getString(DESCRIPTION)); }
	/** @see #documentation() */
	public String documentation (String s) { return((String) set(DESCRIPTION, s)); }
	
	/** URL of web-based help file. */
	public String helpURL () { return(getString(HELPURL)); }
	/** @see #helpURL() */
	public String helpURL (String s) { return((String) set(HELPURL, s)); }
	
	/** Ontological elements contained in the ontology.
	* * Use accessors.
	* @see #elements()
	* @see #elements(Vector)
	* @serial
	*/
	protected Vector elements = null;
	/** Ontological elements contained in the ontology.
	* * Default is new Vector().
	* * Vector is very inefficient. Consider Linked List?
	*/
	public Vector elements () {
		if (elements == null) elements = new Vector(); 
		return(elements); 
	} // reader
	/** @see #elements() */
	protected Vector elements (Vector v) { return(elements = v); }
	
	/** List of all ontologies.
	* * Use accessors.
	* @see #allOntologies()
	* @see #allOntologies(Vector)
	*/
	protected static Vector allOntologies = null;
	/** List of all ontologies.
	* * Default is new Vector().
	* * Vector is very inefficient. Consider Linked List?
	*/
	public static Vector allOntologies () {
		if (allOntologies == null) allOntologies = new Vector(); 
		return(allOntologies); 
	} // reader
	/** @see #allOntologies() */
	public static Vector allOntologies (Vector v) { 
		return(allOntologies = v); 
	} // writer 
	
	public Vector flatten () {
		Vector result = new Vector ();
		Object child = null;
		for(Enumeration kids = children(); kids.hasMoreElements();) {
			child = kids.nextElement();
			if(child instanceof Ontology) result = QRGutils.mergeVector(result, ((Ontology) child).flatten());
			else result.addElement(child);
		} // end for
		return(result);
	} // method

	// Methods
	// -----------------------------------------------------------------------
	/** Generates string for id uniquification. */
	public String composeUniqueIdentifier () {
		String result = "VMODEL-ONTOLOGY"; //getClass().getName(); 
		result += "-" + (new Date()).getTime(); 
		result += "-" + (sessionCount++);
		result += "-" + (new Date()).getTime(); 
		result += "-" + hashCode();
		result += "-" + String.valueOf((char) ((Math.random() * 25) + 65));
		return(result);
	} // method 
	
	// Stringification.
	/** Stringifies the object's KeyValuePairs with indentation.
	* @see PList
	* @see KeyValuePair
	*/
	public String stringifyKeyValuePairs (int indent) {
		String result = "";
		int j = 0;
		KeyValuePair candidate = null;
		if (keyValuePairs() != null) {
			j = 0;
			while(j < keyValuePairs().size()) {
				candidate = (KeyValuePair) keyValuePairs().elementAt(j);
				if (! (candidate.value() instanceof Vector)) {
					if (j > 0) result += QRGutils.eol() + QRGutils.space(indent);
					else result += " "; 
					result += candidate;
				} // end if
				j++;
			} // while
		} // end if
		return(result);
	} // method toString
	
	/** Stringifies the object in XML format with indentation. */
	public String stringify (int indent) {
		String result = "";
		KeyValuePair pair = null;
		Object mumble = null;
		final VectorElement grumble = new VectorElement();
		boolean flag = ! (elements().isEmpty());
		// 1. Print XML tag
		result = QRGutils.space(indent) + "<" + "Ontology";//this.getClass().getName();
		// 2. Stringify key-word pairs
		result += stringifyKeyValuePairs(result.length() + 1);
		// 3. Check for contained objects, vectors, and key-values 
		//    that were not printed above.
		for (Enumeration e = keyValuePairs().elements() ; e.hasMoreElements() ;) {
			pair = (KeyValuePair) e.nextElement();
			if ((pair.value() instanceof Vector) &&
					(! ((Vector) pair.value()).isEmpty())) {
				flag = false;
				break;
			} // end if
		} // end for
		int tabs = indent + 3;
		// 4. If there are none, end XML object.
		if (! flag) result += " />" + QRGutils.eol();
		else {
		// 5. Print contained vectors, and key-values not done above.
			result += " >" + QRGutils.eol();
			for (Enumeration e = keyValuePairs().elements() ; e.hasMoreElements() ;) {
				pair = (KeyValuePair) e.nextElement();
				if (pair.value() instanceof Vector) {
					for (Enumeration p = ((Vector) pair.value()).elements() ; p.hasMoreElements() ;) {
						mumble = p.nextElement();
						grumble.set(TAG, pair.key());
						grumble.set(VALUE, mumble);
						result += QRGutils.space(tabs) + grumble.toString();
					} // end for
				} // end if
			} // end for
			// 6. Print contained elements.
			for (Enumeration o = children() ; o.hasMoreElements() ;) {
				mumble = o.nextElement();
				result += ((Ontology.Element) mumble).stringify(tabs);
				result += QRGutils.eol();
			} // end for
			// 7. End XML object.
			result += QRGutils.space(indent) + "</" + "Ontology" /* this.getClass().getName() */ + ">" + QRGutils.eol();
		} // end if 
		return(result);
	} // method
	
	/** Stringifies the object in XML format.
	*   Indentation starts at column 0.
	*/
	public String stringify () { 
		return(stringify(0)); 
	} // method
	
	/** Stringifies the object. */
	public String toString () {
		return(name());
	} // method
	
	/** Finds the ontology with the specified id. 
	*   Otherwise returns null.
	* * This could be made more efficient, especially with a different data
	*   structure, but how many ontologies are there going to be at any one
	*   time?  Probably less than a dozen.
	*/
	public static Ontology.OntInterface find (String id) {
		Ontology.OntInterface result = null;
		Ontology.OntInterface mumble = null;
		if ((id == null) || (id.equalsIgnoreCase(""))) {
			for (Enumeration o = allOntologies().elements() ; o.hasMoreElements() ;) {
				mumble =  (Ontology.OntInterface) o.nextElement();
				if (id.equalsIgnoreCase(mumble.id())) {
					result = mumble;
					break;
				} // end if
			} // end for
		} // end if
		return(result);
	} // method
	
	
	/** Finds the ontological element with the specified id. 
	*   Otherwise returns null.
	* * Recurses into categories.
	*/
	public Ontology.Element findElement (String id) {
		Ontology.Element result = null;
		Ontology.Element mumble = null;
		for (Enumeration o = children() ; o.hasMoreElements() ;) {
			mumble =  (Ontology.Element) o.nextElement();
			if (id.equalsIgnoreCase(mumble.id())) {
				result = mumble;
				break;
			} // end if
		} // end for
		if (result == null) {
			for (Enumeration o = children() ; o.hasMoreElements() ;) {
				mumble =  (Ontology.Element) o.nextElement();
				if (mumble instanceof Ontology.OntInterface) {
					result = ((Ontology.OntInterface) mumble).findElement(id);
					if (result != null) break;
				} // end if
			} // end for
		} // method
		return(result);
	} // method
	
	/** Returns true if the ontological element with the specified id 
	*   exists in the ontology, otherwise returns false.
	* * Uses findElement and so recurses into categories.
	* @see #findElement(String)
	*/
	public boolean contains (String id) {
		return(findElement(id) != null);
	} // method
	
	/** Removes an ontological element from the ontology. */
	public void remove (Ontology.Element a) {
		elements().removeElement(a);
		a.category(null);
		reload(this);
	} // method
	
	public void removeAll ( ) {
		elements(new Vector());
		reload(this);
	} // method
	
	/** Adds an ontological element to the ontology. */
	public void add (Ontology.Element a) {
		if (a.category() != null) {
			Ontology.OntInterface mumble = find(a.category());
			if ((mumble != null) && (mumble instanceof Ontology)) 
				((Ontology) mumble).remove(a);
		} // end if
		elements().addElement(a);
		a.category(id());
		reload(this);
	} // method
	
	// File I/O methods
	// -----------------------------------------------------------------------
	/** Saves the ontology to filename.
	* * Saves in CML and XML formats as well as java serial format.
	*/
	public void save (String filename) {
		String filepath = QRGutils.QRGuserpath() + filename + "." + ONT_SUFFIX;
		exportCML(filename);
		exportXML(filename);
		try {
			FileOutputStream ostream = new FileOutputStream(filepath);
			ObjectOutputStream p = new ObjectOutputStream(ostream);
			p.writeObject(this);
			p.flush();
			ostream.close();
		//  What about ostream? Potential leak in catches.
		} catch (FileNotFoundException e) {
			QRGutils.debug("Ontology.save", " filename=" + filename + " Error: \n" + e);
		} catch (IOException e) {
			QRGutils.debug("Ontology.save", " filename=" + filename + " Error: \n" + e);
		} // try catch
	} // method save

	/** Loads the ontology from filename.
	* * Loads java serial format.
	*/
	public static Ontology.OntInterface load (String filename) {
		Ontology result = null;
//		try {
//			FileInputStream istream = new FileInputStream(QRGutils.QRGuserpath() + filename + "." + ONT_SUFFIX);
//			ObjectInputStream p = new ObjectInputStream(istream);
//			result = (Ontology) p.readObject();
//			istream.close();

			String fname = QRGutils.QRGuserpath() + filename + "." + XML_SUFFIX;
			result = (Ontology) XMLio.load(fname);
//			istream.close();

		//  What about istream? Potential leak in catches.
//		} catch (FileNotFoundException e) {
//			QRGutils.debug("Ontology.load", " Error: \n" + e);
//			QRGutils.debug("Ontology.load", " Error filename=" + filename);
//		} catch (IOException e) {
//			QRGutils.debug("Ontology.load", " Error: \n" + e);
//			QRGutils.debug("Ontology.load", " Error filename=" + filename);
//		} catch (ClassNotFoundException e) {
//			QRGutils.debug("Ontology.load", " Error: \n" + e);
//			QRGutils.debug("Ontology.load", " Error filename=" + filename);
//		} // try catch
		return(result);
	} // method save

	// CMLio.CMLable
	// -----------------------------------------------------------------------
	/** Saves the ontology to filename in CML format. */
	public void exportCML (String filename) {
		CMLio.save(QRGutils.QRGuserpath() + filename + "." + CML_SUFFIX , this);
	} // method
	
	/** Returns the CML command used to represent the object. */
	public String CMLcommand () { return("defontology"); }

	/** Stringifies the object into CML format. */
	public String toCML () {
		String result = ";;; " + /* getClass().getName() */ "Ontology" + ": " + id() + QRGutils.eol();
		result += ";;; " + NAME + ": " + name() + QRGutils.eol();
		result += ";;; " + HELPURL + ": " + helpURL() + QRGutils.eol();
		result += ";;; " + DESCRIPTION + ": " + documentation() + QRGutils.eol();
		result += QRGutils.eol();
		for (Enumeration o = elements().elements() ; o.hasMoreElements() ;) {
			result += ((CMLio.CMLable) o.nextElement()).toCML();
			result += QRGutils.eol() + QRGutils.eol();
		} // end for
		result += ";;; End of " + "Ontology" /* getClass().getName() */ + ": " + id() + QRGutils.eol();
		result += QRGutils.eol();
		return(result);
	} // method
	
	// XMLio.XMLable
	// -----------------------------------------------------------------------
	/** Saves the ontology to filename in XML format. */
	public void exportXML (String filename) {
		XMLio.save(QRGutils.QRGuserpath() + filename + "." + XML_SUFFIX , this);
	} // method
	
	/** Stringify the an XML definition for the object. */
	public String xmlDef () { 
		String result = 	"";
		KeyValuePair mumble = null;
		String classname = this.getClass().getName();
		result += "<!DOCTYPE " + classname + " [\r\n";
		result += "<!ELEMENT " + classname + " EMPTY >\r\n";
		for (Enumeration e = keyValuePairs().elements() ; e.hasMoreElements() ;) {
			mumble = (KeyValuePair) e.nextElement();
			result += "<!ATTLIST " + classname + " " + mumble.key().toUpperCase() + " CDATA #REQUIRED >\r\n";
		} // end for
		result += "]>";
		return(result); 
	} // method
	
	/** Stringify the object into the XML format. */
	public String xmlify () {
		String result = "";
		result = XMLio.commentate(getClass().getName() + ": " + id()) + QRGutils.eol();
		result += XMLio.commentate(NAME + ": " + name()) + QRGutils.eol();
		result += XMLio.commentate(HELPURL + ": " + helpURL()) + QRGutils.eol();
		result += XMLio.commentate(DESCRIPTION + ": " + documentation()) + QRGutils.eol();
		result += stringify();
		result += XMLio.commentate("End of " + getClass().getName() + ": " + id()) + QRGutils.eol();
		result += QRGutils.eol();
		return(result);
	} // method

	/** Initializes slots after object is created by XML loader. */
	public void XMLioObjectInit (Vector attr, Vector elmnt) {
		keyValuePairs(attr);
		elements(elmnt);
		allOntologies().addElement(this);
	} // method
} // class Ontology

// --------------------------------------------------------------------------
// End of File: Ontology.java                 (c) 2003 Northwestern University
