package edu.qrg.utils;

import edu.qrg.utils.KeyValuePair.KVPinterface;
import java.awt.Color;
import java.io.Serializable;
import java.lang.Cloneable;
import java.lang.Double;
import java.lang.Integer;
import java.lang.NumberFormatException;
import java.lang.Object;
import java.lang.String;
import java.util.Enumeration;
import java.util.Vector;

/** <PRE> Sun Java 1.3.1 -Sourcepath QRG/VModel/development/code/utils
* ----------------------------------------------------------------------------
*  2003 Northwestern University Qualitative Reasoning Group Kenneth D. Forbus
* ----------------------------------------------------------------------------
* File name: PList.java
*    System: VModel
*     Since: v0.6
*    Author: Leo C. Ureel II
* ----------------------------------------------------------------------------
*   Class: PList
* Purpose: A collection of key-value pairs whose values are indexed by key.
* Notes:
* * Would it be more efficient to implement this as a HashTable?
* ----------------------------------------------------------------------------
* </PRE>
* @see KeyValuePair
*/
public class PList
		implements PList.PLinterface, Cloneable, Serializable, QRGconstants {
	// Constructors
	// -------------------------------------------------------------------------
	/** General Intialization. */
	public PList() {
	   super();
	} // default constructor PList

	/** Initialize with a list of KeyValuePairs.
	* * The argument is assumed to be a vector containing only KeyValuePairs.
	*/
	public PList (Vector attr) {
		this();
		keyValuePairs(attr);
	} // constructor PList

	// Interfaces
	// ------------------------------------------------------------------------
	/** <PRE>
	* ------------------------------------------------------------------------
	* Interface: PList.PLinterface
	* Purpose: In general, anything implementing this interface should be
	*          acceptable as a PList.
	* ------------------------------------------------------------------------
	* </PRE> */
	public interface PLinterface extends Cloneable, Serializable, QRGconstants {
		// Accessors
		/** Reader Accessor - returns a vector of KeyValuePairs. */
		public Vector keyValuePairs ();
		/** writer Accessor - sets & returns a vector of KeyValuePairs. */
		public Vector keyValuePairs (Vector v);

		// Methods
		/** Adds a new KeyValuePair to the list.
		    Does not check if key is already in list. */
		public void add (String key, Object value);
		/** Removes the KeyValuePair corresponding to arg from the list. */
		public boolean remove (String key);

		/** Return a list of all keys. */
		public Vector keys ();
		/** Return a list of all values in the PList. */
		public Vector values ();

		// Get key values.
		/** Return the value associated with the specified key. */
		public Object get (String key);

		/** Return the value associated with the specified key as a Color. */
		public Color getColor (String key);

		/** Return the value associated with the specified key as an RGB int. */
		public int getRGBColor (String key);

		/** Return the value associated with the specified key as an int. */
		public int getInt (String key);

		/** Return the value associated with the specified key as a double. */
		public double getDouble (String key);

		/** Return the value associated with the specified key as a boolean. */
		public boolean getBoolean (String key);

		/** Return the value associated with the specified key as a String. */
		public String getString (String key);

		/** Return true if the value associated with the specified key
		    contains the specified target string. */
		public boolean containsString (String key, String target);

		/** Return the value associated with the specified key as a Vector. */
		public Vector getVector (String key);

		// Set key values.
		// Do we need more methods for primitives?
		/** Sets the value associated with the specified key. */
		public Object set (String key, Object value);

		/** Sets the value associated with the specified key. */
		public Object set (String key, int value);

		/** Sets the value associated with the specified key. */
		public Object set (String key, boolean value);

		/** Sets the value associated with the specified key. */
		public Object set (String key, double value);

		/** Sets the value associated with the specified key.
		    Colors are converted to RGB int for internal storage. */
		public Object set (String key, Color value);

		// Stringification
		/** Stringification with indentation. */
		public String stringifyKeyValuePairs (int indent);
		/** Stringification with indentation. */
		public String toString (int indent);
	} // class PList

	// Properties & Accessors
	// ------------------------------------------------------------------------
	/** Storage for the list of key-value pairs.
	* * Property could be a hash table?
	* * Use the accessors.
	* @see #keyValuePairs()
	* @see #keyValuePairs(Vector)
	* @serial
	*/
	protected Vector keyValuePairs = null;
	/** Accessor for the list of key-value pairs.
	* Default: Empty Vector.
	* @see #keyValuePairs
	*/
	public Vector keyValuePairs () {
		if (keyValuePairs == null) keyValuePairs = new Vector();
		return(keyValuePairs);
	} // reader
	/** @see #keyValuePairs */
	public Vector keyValuePairs (Vector v) { return(keyValuePairs = v); }

	/** @deprecated Use keyValuePairs().
	    @see #keyValuePairs() */
	public Vector plist () { return(keyValuePairs()); }
	/** @deprecated Use keyValuePairs(Vector).
	    @see #keyValuePairs(Vector) */
	public Vector plist (Vector v) { return(keyValuePairs(v)); }

	//  Methods
	// -----------------------------------------------------------------------
	/** Returns a Vector of all keys in the PList.
	* Default: Empty Vector.
	*/
	public Vector keys () {
		Vector result = new Vector();
		KeyValuePair.KVPinterface candidate = null;
		for (Enumeration e = keyValuePairs().elements() ; e.hasMoreElements() ;) {
			candidate =  (KeyValuePair.KVPinterface) e.nextElement();
			result.addElement(candidate.key());
		} // end for
		return(result);
	} // method

	/** Returns a Vector of all values in the PList.
	* Default: Empty Vector.
	*/
	public Vector values () {
		Vector result = new Vector();
		KeyValuePair.KVPinterface candidate = null;
		for (Enumeration e = keyValuePairs().elements() ; e.hasMoreElements() ;) {
			candidate =  (KeyValuePair.KVPinterface) e.nextElement();
			result.addElement(candidate.value());
		} // end for
		return(result);
	} // method

	/** Returns the KeyValuePair whose key matches arg.
	* Returns null if key is not found.
	* * Why is this not in the interface?
	* @see KeyValuePair
	*/
	public KeyValuePair.KVPinterface findKey (String key) {
		KeyValuePair.KVPinterface result = null;
		KeyValuePair.KVPinterface candidate = null;
		for (Enumeration e = keyValuePairs().elements() ; e.hasMoreElements() ;) {
			candidate =  (KeyValuePair.KVPinterface) e.nextElement();
			if (key.equalsIgnoreCase(candidate.key())) {
				result = candidate;
				break;
			} // end if
		} // end for
		return(result);
	} // method findKey

	/** Copies the key-value pairs of another PList. */
	public PList copyKeyValuePairs (PList source) {
		KeyValuePair.KVPinterface result = null;
		KeyValuePair.KVPinterface candidate = null;
		for (Enumeration e = source.keyValuePairs().elements() ; e.hasMoreElements() ;) {
			candidate =  (KeyValuePair.KVPinterface) e.nextElement();
			if (candidate.value() instanceof String) {
				set(new String(candidate.key()), new String((String) candidate.value()));
			} else if (candidate.value() instanceof Vector) {
				Vector newValue = new Vector();
				Object item = null;
				for(Enumeration items = ((Vector) candidate.value()).elements(); items.hasMoreElements();) {
					item = items.nextElement();
					if (item instanceof String) newValue.addElement(new String ((String) item));
					else newValue.addElement(item);
				} // end for
				set(new String(candidate.key()), newValue);
			} else set(new String(candidate.key()), candidate.value());
		} // end for
		return(this);
	} // method findKey

	/** Adds a KeyValuePair to the PList.
	* * Not in interface.
	* @deprecated Use add(String key, Object value)
	* @see KeyValuePair
	* @see #add(String, Object)
	*/
	public void add (KeyValuePair.KVPinterface pair) {
		keyValuePairs().addElement(pair);
	} // method

	/** Adds a KeyValuePair to the PList.
	*/
	public void add (String key, Object value) {
		keyValuePairs().addElement(new KeyValuePair(key, value));
	} // method

	/** Removes a KeyValuePair from the PList.
	* * Not in interface.
	* @deprecated Use remove(String key)
	* @see KeyValuePair
	* @see #remove(String)
	*/
	public boolean remove (KeyValuePair.KVPinterface pair) {
		return(keyValuePairs().removeElement(pair));
	} // method

	/** Removes the KeyValuePair associated with key from the PList. */
	public boolean remove (String key) {
		boolean result = false;
		KeyValuePair.KVPinterface candidate = null;
		candidate = findKey(key);
		if (candidate != null)
			result = keyValuePairs().removeElement(candidate);
		return(result);
	} // method set

	/** @deprecated Use remove(String). */
	public boolean removeKey (String key) {
		return(remove(key));
	} // method

	// Get key values.
	/** Return the value associated with key or null if not found. */
	public Object get (String key) {
		Object result = null;
		KeyValuePair.KVPinterface candidate = null;
		candidate = findKey(key);
		if (candidate != null) { result = candidate.value(); }
		return(result);
	} // method

	/** Return the value associated with key as a Color.
	* Defaults to BLACK if key is not found.
	* @see QRGconstants#BLACK
	*/
	public Color getColor (String key) {
		Color result = BLACK;
		Object candidate = get(key);
		String value = (String) candidate;
		if (candidate != null) {
			if (candidate instanceof Color) result = (Color) candidate;
			else {
				value = value.trim();
				if (value==null || value.equals("")) result = Color.black;
				else {
					try {
						int foo = Integer.parseInt(value);
						result = new Color(foo);
					} catch (NumberFormatException e) {
						result = QRGutils.matchColor(value);
					} // end try/catch
				} // end if else
			} // if value
		} // if candidate
		return result;
	} // method

	/** Return the value associated with key as a RGB Color int.
	* Defaults to BLACK if key is not found.
	* @see QRGconstants#BLACK
	*/
	public int getRGBColor (String key) {
		int result = 0;
		Color color = getColor(key);
		if (color != null) result = color.getRGB();
		return(result);
	} // method

	/** Return the value associated with key as an int.
	* Defaults to 0 if key is not found.
	*/
	public int getInt (String key) {
		int result = 0;
		String candidate = (String) get(key);
		if (candidate != null) {
			candidate = candidate.trim();
			if (candidate == null || candidate.equals("")) result = 0;
			else {
				try { result = Integer.parseInt(candidate); }
				catch(NumberFormatException e) { }
			} // if
		} // if
		return result;
	} // method

	/** Return the value associated with key as an double.
	* Defaults to 0.0 if key is not found.
	*/
	public double getDouble (String key) {
		double result = 0.0;
		String candidate = (String) get(key);
		if (candidate != null) {
			candidate = candidate.trim();
			try {result = Double.valueOf(candidate).doubleValue();}
			catch(NumberFormatException e) {}
		} // if value
		return(result);
	} // method

	/** Return the value associated with key as a boolean.
	* Defaults to false if key is not found.
	*/
	public boolean getBoolean (String key) {
		boolean result = false;
		String value = getString(key);
		if (value != null) {
			value = value.trim();
			result = value.equalsIgnoreCase("TRUE");
		} // if value
		return(result);
	} // method

	/** Return the value associated with key as a String.
	* Defaults to "" if key is not found.
	*/
	public String getString (String key) {
		String result = "";
		String value = (String) get(key);
		if (value != null) {
			value = value.trim();
			result = QRGutils.replaceEscapeChars(value);
		} // if value
		return(result);
	} // method

	/** Return true if the value associated with key
	* contains the target string.
	*/
	public boolean containsString (String key, String target) {
		boolean result = false;
		String value = getString(key);
		result =  (value.indexOf(target)>=0);
		return(result);
	} // method getDouble

	/** Return the value associated with key as a Vector.
	* Defaults to new Vector() if key is not found.
	*/
	public Vector getVector (String key) {
		Vector result = null;
		Vector value = (Vector) get(key);
		if (value != null)
			result = value;
		else {
			result = new Vector();
			set(key, result);
		} // end if
		return(result);
	} // method getDouble

	// Set key values.
	/** Associates the specified value with the specified key. */
	public Object set (String key, Object value) {
		Object result = null;
		KeyValuePair.KVPinterface candidate = findKey(key);
		if (candidate != null) result = candidate.value(value);
		else {
			add(key, value);
			result = value;
		} // end if else
		return(result);
	} // method set

	/** @see #set(String, Object) */
	public Object set (String key, int value) {
	   return(set(key, "" + value));
	} // method set

	/** @see #set(String, Object) */
	public Object set (String key, boolean value) {
	   return(set(key, "" + value));
	} // method set

	/** @see #set(String, Object) */
	public Object set (String key, double value) {
	   return(set(key, "" + value));
	} // method set

	/** Color values are stored as RGB int.
	    @see #set(String, Object) */
	public Object set (String key, Color value) {
	   return(set(key, "" + value.getRGB()));
	} // method set

	// Stringification.
	// -----------------------------------------------------------------------
	/** Stringify all KeyValuePairs with the specified indentation. */
	public String stringifyKeyValuePairs (int indent) {
		String result = "";
		int j = 0;
		if (keyValuePairs() != null) {
			j = 0;
			while(j < keyValuePairs().size()) {
				if (j > 0) result += QRGutils.eol() + QRGutils.space(indent);
				else result += " ";
				result += (KeyValuePair) keyValuePairs().elementAt(j);
				j++;
			} // while
		} // end if
		return(result);
	} // method toString

	/** Stringification with indentation. */
	public String toString (int indent) {
		String result = "";
		int i = indent, j = 0;
		String className = this.getClass().getName();
		className = QRGutils.rawClassName(className);  // Added TRH
		result = QRGutils.space(i) + "<" + className;
		result += stringifyKeyValuePairs(result.length() + 1);
		result += " />" + QRGutils.eol();
		return(result);
	} // method toString

	/** Stringify with initial indent = 0. */
	public String toString () {
		return(toString(0));
	} // method toString
} // class PList

// ---------------------------------------------------------------------------
// End of File: PList.java                       2003 Northwestern University
